
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
bmp2arr.cpp
Bitmap to C array conversion program
Niall Murphy 30/1/1999 http://www.iol.ie/~nmurphy
nmurphy@iol.ie


If this program is not mapping the colors in a way
suitable for your target system, then edit the 
getMappedColor function. If you require a version
of this program modified for your application, then
contact Niall Murphy.
*/

/*
Header structures from Charles Petzold's article 
in PC Magazine (vol.10(10), pp365-370).
*/
typedef struct tagBITMAPFILEHEADER
{
        unsigned int  bfType;                   // "BM" or 0x4D42
        unsigned long bfSize;                   // Size of file in bytes
        unsigned int  bfReserved1;              // Set to 0
        unsigned int  bfReserved2;              // Set to 0
        unsigned long bfOffBits;                // Offset in file where
                                                // the bits begin
} BITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER
{
        unsigned long biSize;                   // Size of the structure
        unsigned long biWidth;                  // Width in pixels
        unsigned long biHeight;                 // Height in pixels
        unsigned int  biPlanes;                 // # of color Planes: Set to 1
        unsigned int  biBitCount;               // Color bits per pixel
        unsigned long biCompression;            // Compression Scheme
        unsigned long biSizeImage;              // Number of bitmap bytes
        unsigned long biXPelsPerMeter;          // Horizontal Resolution
        unsigned long biYPelsPerMeter;          // Vertical Resolution
        unsigned long biClrUsed;                // Number of colors used
        unsigned long biClrImportant;           // Important colors

} BITMAPINFOHEADER;

#define BMP_OK -1
#define BMP_BadDriver 0
#define BMP_TooBig 1
#define BMP_WrongVideo 2
#define BMP_Compression 3

/*
Global headers filled by accessBMP and read elsewhere.
*/

BITMAPFILEHEADER BMPHeader;
BITMAPINFOHEADER BMPInfo;

/*
This colorMap  must be hand edited for the conversion to give us the
grey shades that we want.
Note that the in the  4-bit case only the firt 16 entries in the map
will be used.
If the program were adapted to reduce 256 color BMPs to 16 color arrays
then the rest of this would be used.
*/
char colorMap[256];
/*
38400L is 320*240/2, the size of a quarter VGA screen and is
our maximum bitmap size.
*/
char fileBuffer[38400L];

#define BUF_LEN 100

static int accessBMP (char *);                  // func to check BMP file 
static void processBMP(char *BMPFile, char *NDMFile);

/*
This function initialises the array which maps VGA colors to
the values you want to use in the conversion.
This version maps the black, white and two shades of gray,
and all other colors are tagged illegal.
*/
void mapInit(void)
{
    for (int i=0; i < 256; i++)
    {
        colorMap[i] = -1;
    }
    colorMap[15] = 15; /* WHITE */
    colorMap[0] = 0; /* WHITE */
    colorMap[7] = 7; /* DARK GRAY */
    colorMap[8] = 8; /* DARK GRAY */
   
      
}

/*
This function allows access to the colorMap. The array access
could be replaced with some other form of conversion such as
a formula or readig a file.
*/
char getMappedColor(int colorFromBMP)
{
//    return colorMap[colorFromBMP];
	/*
	I found that this +1 mapping worked for the
	bitmaps that I coverted from 24bit bitmaps in
	Photoshop to 16 bit greyscale. The mapping you
	use may depend on the sourse of your bitmap and
	the colors on your target device.
	*/
		return (colorFromBMP+1) % 16;
}

void main(int argc, char *argv[])
{
    int x;
    
    if (argc != 3)
    {
		  printf("Usage: bmp2arr infile.bmp outfile.c\n");
        exit(-1);
    }
    if (strcmp(argv[1], argv[2]) == 0)
    {
        printf("input and output can not have the same name\n");
        exit(-1);
    }


    x  = accessBMP(argv[1]);                // Check out the BMP image to
                                            // see if it's legitimate   

    printf("\nFILE: %s",argv[1]);

	 printf("\nType of BMP: ");
						 if (BMPInfo.biBitCount == 1) printf("Monochrome");
						 else if (BMPInfo.biBitCount == 4) printf("16-Color");
						 else if (BMPInfo.biBitCount == 8) printf("256-Color");
                   else printf("24-Bit RGB");
    if (BMPInfo.biCompression==0) printf("\nCompression Scheme: NONE");
    else                          printf("\nCompression Scheme: %u",BMPInfo.biCompression);
    printf("\nStructure Size: %u",BMPHeader.bfSize-BMPHeader.bfOffBits);
    printf("\nWidth: %u",BMPInfo.biWidth);
    printf("\nHeight: %u",BMPInfo.biHeight);
    printf("\nBits per pixel: %u",BMPInfo.biBitCount);
    printf("\nBitmap Bytes: %u",BMPInfo.biSizeImage);
    printf("\n");

    switch(x)
    {
				case BMP_OK:            printf("\nProceeding to Convert.\n\n");
												break;
				case BMP_WrongVideo:    printf("\nThis program can only");
												printf("\n\taccomodate 16-Color BMP's.");
												break;
				case BMP_Compression:   printf("\nBMP file was saved in a Compressed");
												printf("\n\tformat which is not supported.");
                                    break;
            default:                printf("\n* UNKNOWN ERROR *");
                                    printf("\n**** ABORT ****\n\n");
                                    break;
    }

 
    
    mapInit();
    
	 processBMP(argv[1], argv[2]);
}
    
    
    
    

/*
This function reads the header information which is later
used by the convertion routine.
*/
int accessBMP(char *BMPFile)
{
		  FILE *Fn;
        int errorcode = BMP_OK;

        Fn = fopen(BMPFile,"rb");               // Open up BMP image file     

                                                // Read in BMP header
		  fread(&BMPHeader,sizeof(BMPHeader),1,Fn);
        fread(&BMPInfo,sizeof(BMPInfo),1,Fn);
		  fclose(Fn);

		  if (BMPInfo.biBitCount != 4)
        {
				 errorcode = BMP_WrongVideo;
        }

                                                // Error: image was compressed 
		  if (BMPInfo.biCompression != 0)
        {
				errorcode = BMP_Compression;
        }

        return(errorcode);
}

/*
This routine actualy performs the conversion
*/
void processBMP(char *BMPFileName, char *NDMFileName)
{
	 FILE *Fn, *ndmFile;                          // Fn: BMP file;
	 int x, y;
	 unsigned int width, height;
	 char Color;
	 char rootName[BUF_LEN];

	 int bytesLeftOver;
	 int i;

	 Fn = fopen(BMPFileName,"rb");               // Open BMP file


	 ndmFile = fopen(NDMFileName,"w");
														  // Jump to image data in
	 fseek(Fn,BMPHeader.bfOffBits,SEEK_SET); // BMP file

 // truncate the filename to get the data structure name
	 // After the loop stringBuf will be the root name of the file.
	 // Note that this can not handle directories
	 strcpy(rootName, NDMFileName);
	 char *bufPtr = rootName;
	 while (*bufPtr)
	 {
		  if (*bufPtr == '.')
		  {
				*bufPtr = '\0';
		  }
		  bufPtr++;
	 }



	 if (BMPInfo.biWidth % 2 != 0)
	 {
		printf("Sorry, but I can not handle conversion unless the width is an even number");
        exit(-1);
    }
    /*
    The width needs to be the width in pixles. We then have to figure out
    how many  bytes are left over.
    */
	 width = BMPInfo.biWidth;
	 height = BMPInfo.biHeight;

    /*
    You would expect width/2 butes, but some rounding occurs and we need to throw
    away the extra bytes at the end of each line. These bytes are in the .bmp,
    but are never displayed so we are not losing any information here.
    */
    bytesLeftOver = ((((4*width)+31)/32) * 4) - (width/2);
    
    /*
    bitmaps.h will define BitmapData which will be used later bu the source code
    The whole structure is filled out here to avoid header file worries.
    */
    fprintf(ndmFile, "#include \"bitmaps.h\"\n");
     
    fprintf(ndmFile, "const static unsigned char GSB_raw_%s[] = {\n", rootName);
    


        y = height-1;
                                                // Read in & display BMP
                                                // Note: BMP's are stored in
                                                // bottom-up fashion, so we'll
                                                // draw from the bottom to the top
                                                // Also, since the program is setup
                                                // only for 16-Color BMP's, there
                                                // are 4 bits for each pixel
        do
        {
            for (x=0; x < width; x+=2)
            {
                if (fread(&Color,sizeof(Color),1,Fn) != 1)
                {
                    printf("Unxpectd EOF");
                    exit(-1);
                }
					 fileBuffer[(long)y*((long)width/2) + x/2] = Color;
            }                 
            /*
            Skip bytesLeftOver bytes at the end of each line.
            */

            for (i=0; i<bytesLeftOver; i++)
            {
                if (fread(&Color,sizeof(Color),1,Fn) != 1)
                {
                        printf("Unxpectd EOF at end of line");
                        exit(-1);
                }
                          
                if (Color != 0)
                {
                        printf("EOL non-zero");
                        /*exit(-1); */
                }
            }
                       
         } while (y-- > 0);
                                                
                    
         for(y=0; y<height; y++)
         {
            for(x=0; x< width; x+=2)
            {
                int numIn;
					 Color = fileBuffer[(long)y*((long)width/2)+x/2];
                numIn = (Color >> 4) & 0x0F;
                if (getMappedColor(numIn) == -1)
                {
                    printf("umapped color in input :%d\n" ,numIn);
                    exit(-1);
                }
                unsigned char twoPixels;
                twoPixels = getMappedColor(numIn) << 4;
                numIn = Color & 0x0F;
                if (getMappedColor(numIn) == -1)
                {
                    printf("umapped color in input :%d\n" ,numIn);
                    exit(-1);
                }
                twoPixels |= getMappedColor(numIn);
                fprintf(ndmFile, "%d", (int)twoPixels);
                /*
                It is important to skip the comma on the last byte.
                */
                if (! ((y == height-1 ) && (x ==  width - 2)) )
                {
                    fprintf(ndmFile, ", ");
                }
                if ((x != 0) &&((x % 20) == 0))
                {
                    fprintf(ndmFile, "\n");
                }
            } /* end for x */
            fprintf(ndmFile, "\n\n\n");
         } /* end for y */
              

         fprintf(ndmFile,"};\n");
        // GB_ stands for global bitmap
        fprintf(ndmFile, "const BitmapData GB_%s = {\n", rootName);
        fprintf(ndmFile, "%d, %d, \n ", width, height);
        fprintf(ndmFile, "GSB_raw_%s }; \n", rootName);
        
                                              

         fclose(ndmFile);
         fclose(Fn);
         printf("\n Conversion complete");
}

