
#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <strstrea.h>
#include <assert.h>
/*
bmp2arr.cpp
BDF to C array conversion program
Niall Murphy 30/1/1999 http://www.iol.ie/~nmurphy
*/
/*
This program coverts from .BDF format to an array of arrays, which 
represnts an array of character definitions. The C code
produced can then be linked into a program which does not support
fonts which are stored in files seperate to the executable.

One restriction of this program is that the .BDF
file must store each record on a seperate line. I have not come
across any applications that do not do this, but they ay be out there
since the standard does not insist on a newline character after each
record.

A more important retriction is that this program does not support
doublebyte characters - so no Kanji. Contact Niall Murphy (nmurphy@iol.ie)
if you require this.
*/

#define MAX_NAME_LEN 100
/*
We are restricting ourselves to single byte accss here, though extending
for doublebyte access should be fairly straightforward. The main problem
is that the array produced becomes enormous.
*/
#define MAX_CHAR_SET_SIZE 256

/*
This differs from the structure used in the rendering code, in
that the data is a copy here, not a pointer. The reason for
allocating the space here is thet the data is constructed here,
not simply refered to as it is in the rendering code.
*/
typedef struct  {
	int offsetX;    /* baseline of font character */
	int offsetY;
	int width;       /* width of char in pixels (bits) */
	int height;      /* height of char in pixels (bits) */
   unsigned char dataName[MAX_NAME_LEN];   /* raw bitmap data */
} CharData;

CharData  FS_charArray[MAX_CHAR_SET_SIZE];


void dobitmaps(ifstream &inFile, ofstream &outFile, char *fontname);
void constructArray(ifstream &inFile, char *fontname);
void outputArray(ofstream &outFile, char *fontName);



void main(int argc, char *argv[])
{
	if (argc != 4)
    {
		cout << "Usage: fconv infile.bdf outfile.c fontname" << endl;
        exit(0);
    }
	ifstream inFile(argv[1]);
	ofstream outFile(argv[2]);

    if (!inFile)
    {
    	cerr << "Failed to open" << argv[1] << endl;
        exit(0);
    }

    if (!outFile)
    {
    	cerr << "Failed to open" << argv[2] << endl;
        exit(0);
    }

	 dobitmaps(inFile, outFile, argv[3]);
    // restart the input file to process all of the characters again
    // for the array of structures part.
    inFile.close();
    inFile.open(argv[1]);
	constructArray(inFile, argv[3]);
	outputArray(outFile, argv[3]);

}


void dobitmaps(ifstream &inFile, ofstream &outFile, char *fontName)
{
	unsigned char line[200];
	unsigned char name[200];
	unsigned int encoding;
	int charCount=0;
	int width, height, offsetX, offsetY;
	int i,j;
	 int widthInBytes;

    outFile << hex;
    while (inFile.getline(line, 200))
    {
        // find next start of char
		  while (strncmp(line, "STARTCHAR", 9) != 0)
        {

				if (strncmp(line, "ENDFONT", 7) == 0)
				{
                cout << " Character bitmap conversion complete for " << dec<< charCount << " characters" << endl;
                return;
            }
            inFile.getline(line, 200);
        }
        // extract decimal value of the character
        inFile.getline(line, 200);
        if (strncmp(line, "ENCODING", 8) != 0)
        {
            cerr << "bad file format, missing ENCODING" << endl;
        }
        
		  istrstream lineEncoding(line);
		  lineEncoding >> name >> encoding;

        // find next start of bits
		while (strncmp(line, "BBX", 3) != 0)
		{
			assert(inFile.getline(line, 200));
		}
        
        istrstream lineBBX(line);
        lineBBX >> name >> width >> height >> offsetX >> offsetY;
        
        // widthInBytes is the number of bytes required to describe one row of the character
        widthInBytes = ((width-1) / 8) + 1;

		// Switch from hex to decimal.
		cout <<  dec ;
    	// title of character
    	outFile << "static char " << fontName << "_" << encoding << "_bits[] =" << endl << "{" << endl;

		// find next start of bits
		while (strncmp(line, "BITMAP", 6) != 0)
		{
			inFile.getline(line, 200);
		}

    	// convet each hex number to array format
        // One line of the .bdf file will correspont to one line of 
        // dots in the character, though the number of bytes in this
        // line may vary.
    	inFile.getline(line, 200);

		outFile << hex;
		for(i=0; i<height*widthInBytes; i+=widthInBytes)
    	{
			// loop j from the first to last byte of a given line.
            // Each byte on a line covers 8 bits of data
            for(j=0; j<widthInBytes; j++)
            {
                // two characters represent one hex byte
					outFile << "0x" << line[j*2] << line[j*2+1] ;
					// only do the comma if we are not at the last byte.
					if ((i+j)<(height*widthInBytes)-1)
					{
						outFile <<",";
					}
				}
			outFile << endl;
			inFile.getline(line, 200);
    	}

    	outFile << "};" << endl;
    	charCount++;
    } // end while


} // end map


void constructArray(ifstream &inFile, char *fontName)
{
	unsigned char line[200];
	unsigned char name[200];
    /*
    fieldName is generally used to stream in the name at the start
    of the line (record), but generally does not need to be examined,
    it is just skipped over to get to the data
    */
	unsigned char fieldName[100];
    unsigned int encoding;
    int charCount=0;
    int width, height, offsetX, offsetY;
    int i;

    // Initialise the char array to contain no characters
    for(i = 0; i < MAX_CHAR_SET_SIZE; i++)
	 {
		  // the offset and size are set to zero so we do not
		  // get garbage when we output the unused characters.
		  // The string is set to empty to tag this char as
		  // unused.
		  FS_charArray[i].offsetX = 0;
		  FS_charArray[i].offsetY = 0;
		  FS_charArray[i].width = 0;
		  FS_charArray[i].height = 0;
		  FS_charArray[i].dataName[0] = '\0';
    }
    
    while (inFile.getline(line, 200))
    {
        // find next start of char
		  while (strncmp(line, "STARTCHAR", 9) != 0)
		  {
				inFile.getline(line, 200);
				if (strncmp(line, "ENDFONT", 7) == 0)
				{
					cout << " Font conversion complete for " << charCount << " characters" << endl;
						  return;
				}
		  }


        // extract decimal value of the character
        inFile.getline(line, 200);
        if (strncmp(line, "ENCODING", 8) != 0)
        {
            cerr << "bad file format, missing ENCODING" << endl;
        }
        istrstream lineEncoding(line);
		lineEncoding >> fieldName >> encoding;

        // check that this character is not a repeat
        if (FS_charArray[encoding].dataName[0] != '\0')
        {
				cerr << "Warning: repeated character " << encoding << ", overwritting earlier one" << endl;
        }
            
		// find next start of bits
		while (strncmp(line, "BBX", 3) != 0)
		{
			assert(inFile.getline(line, 200));
		}
		  istrstream lineBBX(line);
        lineBBX >> fieldName;
        lineBBX >> FS_charArray[encoding].width;
        lineBBX >> FS_charArray[encoding].height;
		  lineBBX >> FS_charArray[encoding].offsetX;
        lineBBX >> FS_charArray[encoding].offsetY;
        
        // construt the name of the data
        ostrstream nameStream(FS_charArray[encoding].dataName, MAX_NAME_LEN);
    	nameStream << fontName << "_" << hex << encoding << dec << "_bits" << ends;
        
        charCount++;
	 }
} // end map


void outputArray(ofstream &outFile, char *fontName)
{
	 int charIndex;
    
	 outFile << endl;

	outFile << endl <<  "CharInfo "  << fontName << "[" << dec<< MAX_CHAR_SET_SIZE << "] = \n{" << endl;
    
	 for(charIndex=0; charIndex < MAX_CHAR_SET_SIZE; charIndex++)
    {
        

		outFile << " { ";
		outFile << FS_charArray[charIndex].offsetX << ", ";
		outFile << FS_charArray[charIndex].offsetY << ", ";
		outFile << FS_charArray[charIndex].width << ", ";
		outFile << FS_charArray[charIndex].height << ", ";
		if (FS_charArray[charIndex].dataName[0] != '\0')
		{
			outFile << FS_charArray[charIndex].dataName;
		}
		else
		{
			outFile << "NULL";
		}

		outFile << " }";
        
        // only do a comma if we are not at the last character.
        if (charIndex != MAX_CHAR_SET_SIZE-1)
        {
            outFile << ",";
        }
        outFile << endl;

    }
    outFile << endl << "};" << endl;
}


